#include <allegro.h>
#include <alleggl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <loadpng.h>
#include <string.h>

static int windowed = 0;
static int frames = 0;
static volatile int ticks = 0;
static const int left = KEY_LEFT;
static const int right = KEY_RIGHT;
static const int up = KEY_UP;
static const int down = KEY_DOWN;

static int babies_left = 0;
static int scrollx = 0, scrolly = 0;
static int level = 0;

static int startx = 0, starty = 0;

enum {GAME, LEVEL, QUIT};
static int state = GAME;

typedef struct Being Being;
struct Being
{
    float x, y;
    int ix, iy;
    int frame, dir;
    int id;
    int layer;

    int home;
    float dx, dy;
    int follow;
    int slowness;
};

static int beings_count = 1;
Being beings[16 * 16];
int is[16 * 16];

static int id_numbers = 1;
static int id_crate = 2;
static int id_troll = 5;
static int id_jumper = 69;
static int id_baby = 133;

typedef struct Quad Quad;
struct Quad
{
    BITMAP *pixmap;
    GLuint texture;
    int ox, oy;
    int l, t, r, b;
};

static Quad quads[16 * 16];
static int map[16 * 16 * 16 * 16 * 16];

static int quads_count = 1;

int leveldown = 0;
int goinghome = 0;
SAMPLE *s_scared, *s_laugh, *s_sound, *s_organ, *s_step, *s_boing;

static int being_move(Being *self, float dx, float dy);

static void ticker(void)
{
    ticks++;
}

static int pixelx(int ix, int iy)
{
    return ix * 16 - iy * 16;
}

static int pixely(int ix, int iy)
{
    return ix * 8 + iy * 8;
}

static int isox(int px, int py)
{
   return (py + py + px) >> 5;
}

static int isoy(int px, int py)
{
   return (py + py - px) >> 5;
}

static inline int *pos(int x, int y, int z)
{
    return map + (y & 255) * 16 * 16 * 16 + (x & 255) * 16 + z;
}

static void read_quads(char const *picname, char const *txtname)
{
    int i = quads_count;
    BITMAP *pixmap = load_bitmap(picname, NULL);
    GLuint texture = allegro_gl_make_texture(pixmap);
    PACKFILE *file = pack_fopen(txtname, "r");
    int s = file_size(txtname);
    char *buf = malloc(s + 1);
    pack_fread(buf, s, file);
    buf[s] = '\0';
    pack_fclose(file);
    char *ptr = buf;
    while (*ptr)
    {
        while (*ptr && *ptr != ' ')
            ptr++;
        if (*ptr == ' ')
            ptr++;
        quads[i].pixmap = pixmap;
        quads[i].texture = texture;
        quads[i].ox = strtol(ptr, &ptr, 10);
        quads[i].oy = strtol(ptr, &ptr, 10);
        quads[i].l = strtol(ptr, &ptr, 10);
        quads[i].t = strtol(ptr, &ptr, 10);
        quads[i].r = strtol(ptr, &ptr, 10);
        quads[i].b = strtol(ptr, &ptr, 10);
        is[i] = 0;
        if (i >= id_troll && i < id_troll + 64)
        {
            quads[i].oy += 16;
            is[i] = id_troll;
        }
        else if (i >= id_jumper && i < id_jumper + 64)
            quads[i].oy += 16;
        else if (i >= id_baby && i < id_baby + 64)
        {
            quads[i].oy += 16;
            is[i] = id_baby;
        }
        while (*ptr && *ptr != '\n')
            ptr++;
        if (*ptr == '\n')
            ptr++; 
        i++;
    }
    quads_count = i;
}

static Being *being_new(int id, int x, int y, int layer)
{
    int i = beings_count;
    Being *self = &beings[i];
    memset(self, 0, sizeof *self);
    self->id = id;
    self->x = self->ix = x * 16;
    self->y = self->iy = y * 16;
    self->layer = layer;
    if (beings_count < 255)
        beings_count++;
    being_move(self, 0, 0);
    if (id == id_baby)
        babies_left++;
    return self;
}

static void create_room(int xp, int yp, int w, int h, int corridor, int light,
    int flor, int slugs, int babies)
{
    int x, y;
    for (y = 0; y < h; y++)
    {
        for (x = 0; x < w; x++)
        {
            if ((x == 0 && !(corridor & 1)) ||
                (y == 0 && !(corridor & 2)) ||
                (x == w - 1 && !(corridor & 4)) ||
                (y == h - 1 && !(corridor & 8)))
            {
                if (*pos(xp + x, yp + y, 0) == 0)
                {
                    *pos(xp + x, yp + y, 0) = 0;
                    *pos(xp + x, yp + y, 1) = id_crate;
                }
            }
            else
            {
                if (*pos(xp + x, yp + y, 0) != id_crate + 2)
                {
                    *pos(xp + x, yp + y, 0) = flor;
                    *pos(xp + x, yp + y, 1) = 0;
                }
            }
            int l = *pos(xp + x, yp + y, 6);
            if (l < light)
                *pos(xp + x, yp + y, 6) = light;
        }
    }
    int i;
    for (i = 0; i < slugs; i++)
    {
        x = rand() % (w - 2);
        y = rand() % (h - 2);
        being_new(id_jumper, xp + 1 + x, yp + 1 + y, 2);
    }

    for (i = 0; i < babies; i++)
    {
        x = rand() % (w - 2);
        y = rand() % (h - 2);
        being_new(id_baby, xp + 1 + x, yp + 1 + y, 3);
    }
}

static int fit(int l1, int l2)
{
    if (l1 == l2)
        return 0;
    if (l1 < l2)
    {
        return -(rand() % (l2 - l1 + 1));
    }
    else
    {
        return rand() % (l1 - l2 + 1);
    }
}

static void create_level(void)
{
    int x, y, z;
    for (y = 0; y < 16 * 16; y++)
    {
        for (x = 0; x < 16 * 16; x++)
        {
            for (z = 0; z < 16; z++)
            {
                *pos(x, y, z) = 0;
            }
        }
    }
    beings_count = 1;
    babies_left = 0;

    int orx = 0;
    int ory = 0;
    int orw = 10 + level;
    int orh = 10 + level;

    create_room(0, 0, orw, orh, 0, 255, id_crate + 2, 0, 0);

    startx = orw / 2;
    starty = orh / 2;
    being_new(id_troll, startx, starty, 1);

    int rx, ry, rw, rh;
    int n = 10 + level * 7;
    int rs = 5 + level;
    int corridor_length = 10 + level * 5;
    int i;
    int sl = 1 + level / 3;
    int ba = 1 + level;
    int room = 0;
    int d = rand() % 4;
    for (i = 0; i < n; i++)
    {
        int gotroom = 0;
        int l, b;
        if (room)
        {
            gotroom = 1;
            l = 10 + rand() % rs;
            b = 10 + rand() % rs;
            room = 0;
        }
        else
        {
            d = rand() % 4;
            room = (rand() % 4) == 0;
            if (i == n - 1)
                room = 0;
            l = 8 + rand() % corridor_length;
            b = 4 + rand() % 2;
        }
        switch (d)
        {
            case 0:
                rx = orx + orw - 1;
                ry = ory + fit(orh, b);
                rw = l;
                rh = b;
                break;
            case 1:
                rx = orx + fit(orw, b);
                ry = ory + orh - 1;
                rw = b;
                rh = l;
                break;
            case 2:
                rx = orx - l + 1;
                ry = ory + fit(orh, b);
                rw = l;
                rh = b;
                break;
            case 3:
                rx = orx + fit(orw, b);
                ry = ory - l + 1;
                rw = b;
                rh = l;
                break;
        }
        if (gotroom)
            create_room(rx, ry, rw, rh, 0, 200 + rand() % 100, id_crate, sl, ba);
        else
        {
            int walls = 1 << d;
            if (room)
            {
                walls += 1 << ((d + 2) & 3);
            }
            create_room(rx, ry, rw, rh, walls, 0, id_crate + 1, 0, i == n - 1 ? 1 : 0);
        }
        orx = rx & 255;
        ory = ry & 255;
        orw = rw & 255;
        orh = rh & 255;
    }
}

static void init(void)
{
    loadpng_init();

    install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);

    allegro_gl_set_texture_format(GL_RGBA8);
    set_color_conversion(COLORCONV_NONE);

    read_quads("sheet.png", "sheet.txt");
    read_quads("sheet2.png", "sheet2.txt");

    s_scared = load_sample("scared.wav");
    s_laugh = load_sample("laugh.wav");
    s_sound = load_sample("sound.wav");
    s_organ = load_sample("organ.wav");
    s_step = load_sample("step.wav");
    s_boing = load_sample("boing.wav");

    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, SCREEN_W, SCREEN_H, 0, -1, 1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0, 0, 0, 0);

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    create_level();

}

static inline int clamp(int val)
{
    if (val < -128)
        return -128;
    if (val > 127)
        return 127;
    return val;
}

static int being_move(Being *self, float dx, float dy)
{
    int ix, iy;
    unsigned int sx, sy;
    int r = 0;

    ix = self->ix >> 4;
    iy = self->iy >> 4;
    *pos(ix, iy, self->layer) = 0;

    ix = self->ix >> 4;
    iy = (self->iy + 15) >> 4;
    *pos(ix, iy, self->layer) = 0;

    ix = (self->ix + 15) >> 4;
    iy = self->iy >> 4;
    *pos(ix, iy, self->layer) = 0;

    ix = (self->ix + 15) >> 4;
    iy = (self->iy + 15) >> 4;
    *pos(ix, iy, self->layer) = 0;

    float ox = self->x;
    float oy = self->y;
    int oix = self->ix;
    int oiy = self->iy;

    self->x += dx;
    self->y += dy;
    if (self->x >= 16 * 16 * 16)
        self->x -= 16 * 16 * 16;
    if (self->y >= 16 * 16 * 16)
        self->y -= 16 * 16 * 16;
    if (self->x < 0)
        self->x += 16 * 16 * 16;
    if (self->y < 0)
        self->y += 16 * 16 * 16;

    self->ix = (int)self->x;
    self->iy = (int)self->y;

    int i;
    for (i = 1; i < 4; i++)
    {
        ix = self->ix >> 4;
        iy = self->iy >> 4;
        int c1 = *pos(ix, iy, i) & 255;

        ix = self->ix >> 4;
        iy = (self->iy + 15) >> 4;
        int c2 = *pos(ix, iy, i) & 255;

        ix = (self->ix + 15) >> 4;
        iy = self->iy >> 4;
        int c3 = *pos(ix, iy, i) & 255;

        ix = (self->ix + 15) >> 4;
        iy = (self->iy + 15) >> 4;
        int c4 = *pos(ix, iy, i) & 255;

        if (c1 || c2 || c3 || c4)
        {
            if (i == 1 || i == self->layer)
            {
                self->x = ox;
                self->y = oy;
                self->ix = oix;
                self->iy = oiy;
            }
            if (i == 1)
            {
                if (is[c1] == id_troll || is[c2] == id_troll ||
                    is[c3] == id_troll || is[c4] == id_troll)
                    r = id_troll;
                break;
            }
            else if (i == 3)
            {
                if (is[c1] == id_baby || is[c2] == id_baby ||
                    is[c3] == id_baby || is[c4] == id_baby)
                    r = id_baby;
            }
       }
    }

    int id = self->id + (self->frame >> 5) + self->dir * 8;

    ix = self->ix >> 4;
    iy = (self->iy + 15) >> 4;
    sx = clamp(((int)self->ix - ix * 16) * 8);
    sy = clamp(((int)self->iy - iy * 16) * 8);
    *pos(ix, iy, self->layer) = id + (sx << 8) + (sy << 16);

    ix = (self->ix + 15) >> 4;
    iy = self->iy >> 4;
    sx = clamp((self->ix - ix * 16) * 8);
    sy = clamp((self->iy - iy * 16) * 8);
    *pos(ix, iy, self->layer) = id + (sx << 8) + (sy << 16);

    ix = (self->ix + 15) >> 4;
    iy = (self->iy + 15) >> 4;
    sx = clamp((self->ix - ix * 16) * 8);
    sy = clamp((self->iy - iy * 16) * 8);
    *pos(ix, iy, self->layer) = id + (sx << 8) + (sy << 16);

    return r;
}

static int being(Being *self, float dx, float dy)
{
    int r = 0;
    float d = sqrt(dx * dx + dy * dy);

    if (d)
    {
        self->frame += 8;
        self->frame &= 255;
        self->dir = (int)(3.5 + (AL_PI + atan2(dx, dy)) * 4 / AL_PI) & 7;
        int r1 = being_move(self, sqrt(2) * dx / d, 0);
        int r2 = being_move(self, 0, sqrt(2) * dy / d);
        if (r1 == id_troll || r2 == id_troll)
            r = id_troll;
        else if (r1 == id_baby || r2 == id_baby)
            r = id_baby;
    }
    return r;
}

static void being_distance(Being *self, float x, float y, float *dx_, float *dy_)
{
    static float m = 16 * 16 * 16;
    float dx = x - self->x;
    float dy = y - self->y;
    if (dx > m / 2)
        dx -= m;
    if (dx < -m / 2)
        dx += m;
    if (dy > m / 2)
        dy -= m;
    if (dy < -m / 2)
        dy += m;
    *dx_ = dx;
    *dy_ = dy;
}

static int get_being(int i, int id, float x, float y, float r)
{
    for (; i < beings_count; i++)
    {
        Being *b = &beings[i];
        if (!b->home && b->id == id)
        {
            float dx, dy;
            being_distance(b, x, y, &dx, &dy);
            float q = dx * dx + dy * dy - r * r;
            if (q <= 0)
                return i;
        }
    }
    return 0;
}

static void player(Being *self)
{
    int kx = 0, ky = 0;
    if (key[left])
        kx = -1;
    if (key[right])
        kx = 1;
    if (key[up])
        ky = -1;
    if (key[down])
        ky = 1;

    float dx = kx + ky;
    float dy = ky - kx;

    int r = being(self, dx, dy);
    if (dx || dy)
    {
        if (self->frame == 64 || self->frame == 192)
            play_sample(s_step, 200, 128 - 10 + rand() % 21, 1000, 0);
    }
    if (r == id_baby)
    {
        int bi = 2;
        while ((bi = get_being(bi, id_baby, self->x, self->y, 32)))
        {
            Being *baby = &beings[bi];
            if (!baby->follow)
            {
                baby->follow = 1;
                play_sample(s_laugh, 255, 128, 1000, 0);
            }
            bi++;
        }
    }

    scrollx = (self->x - self->y) - 320;
    scrolly = (self->x + self->y) / 2 - 240;

    int x, y;
    int px = self->ix >> 4;
    int py = self->iy >> 4;
    for (y = -5; y <= 5; y++)
    {
        for (x = -5; x <= 5; x++)
        {
            *pos(px + x, py + y, 7) = 300 - sqrt(x * x + y * y) * 38;
        }
    }

    x = self->ix >> 4;
    y = self->iy >> 4;
    if (*pos(x, y, 0) == id_crate + 2)
    {
        if (babies_left == 0)
        {
            state = LEVEL;
        }
    }
}

static void play_distant_sound(SAMPLE *s, float x, float y, int fr)
{
    Being *troll = &beings[1];
    float dx = x - troll->x;
    float dy = y - troll->y;
    if (dx * dx + dy * dy < 255 * 255)
    {
        int vol = sqrt(dx * dx + dy * dy);
        vol = 255 - vol;
        play_sample(s_boing, vol, 128, fr, 0);
    }
}

static void jumper(Being *self)
{
    if (self->slowness)
    {
        self->slowness--;
        return;
    }
    else
        self->slowness = 1 + rand() % 3;

    Being *troll = &beings[1];

    float dx, dy;
    being_distance(self, troll->x, troll->y, &dx, &dy);
    if (self->home)
    {
        dx = -dx;
        dy = -dy;
    }

    int r = being(self, dx, dy);
    if ((dx || dy) && self->frame == 64)
        play_distant_sound(s_boing, self->x, self->y, 250);
    if (r == id_troll)
    {

    }
    else if (r == id_baby && !self->home)
    {
        int bi = 2;
        while ((bi = get_being(bi, id_baby, self->x, self->y, 32)))
        {
            Being *baby = &beings[bi];
            if (baby->follow)
            {
                baby->follow = 0;
                play_sample(s_scared, 255, 128, 1000, 0);
            }
            bi++;
        }
    }

    int x = self->ix >> 4;
    int y = self->iy >> 4;
    if (*pos(x, y, 0) == id_crate + 2)
    {
        self->home = 1;
    }
}

static void baby(Being *self)
{
    float dx = self->dx, dy = self->dy;
    Being *troll = &beings[1];

    if (self->home)
    {
        if (frames % 120 == 0)
        {
            being_distance(self, startx * 16 + rand() % 10,
                starty * 16 + rand() % 10, &dx, &dy);
            self->dx = dx;
            self->dy = dy;
        }
    }
    else if (self->follow)
    {
        being_distance(self, troll->x, troll->y, &dx, &dy);
        int x = self->ix >> 4;
        int y = self->iy >> 4;
        if (*pos(x, y, 0) == id_crate + 2)
        {
            babies_left--;
            self->home = 1;
            goinghome = 30 * 4;
        }
    }
    else
    {
        if ((frames % 120) == 0)
        {
            self->dx = rand() % 3 - 1;
            self->dy = rand() % 3 - 1;
        }
    }

    being(self, dx, dy);
    if ((dx || dy) && self->frame == 64)
        play_distant_sound(s_boing, self->x, self->y, 1000);
}

static void play_note(SAMPLE *i, int n, int v, int d)
{
    float fr = 16.35159783 * pow (2.0, n / 12.0);
    play_sample(i, v, 128 + rand() % (1 + d) - d / 2, 1000 * fr / 440, 0);
}

static void logic(void)
{
    player(&beings[1]);
    int i;
    for (i = 2; i < beings_count; i++)
    {
        Being *self = &beings[i];
        if (self->id == id_jumper)
            jumper(self);
        if (self->id == id_baby)
            baby(self);
    }
    if (state == LEVEL)
    {
        level++;
        if (level == 10)
        {
            allegro_message("You have won the game!");
            state = QUIT;
        }
        else
        {
            create_level();
            state = GAME;
            leveldown = 15 * 5;
        }
    }
    if (leveldown)
    {
        int f[] = {100, 98, 96, 94};
        if (leveldown % 15 == 0)
        {
            int note = f[5 - leveldown / 15];
            play_note(s_sound, note, 255, 0);
        }
        leveldown--;
    }
    if (goinghome)
    {
        if (goinghome % 30 == 0)
        {
            int note = goinghome / 30;
            play_note(s_organ, note == 4 ? 60 : 72, 255, 0);
        }
        goinghome--;
    }
}

static void subquad(int i, int s)
{
    glBindTexture(GL_TEXTURE_2D, quads[i].texture);
    glBegin(GL_TRIANGLE_FAN);

    GLfloat w = 16 * 16;
    GLfloat h = 16 * 16;

    GLfloat l = quads[i].l + (s & 3) * 8;
    GLfloat t = quads[i].t + (s >> 2) * 8;
    GLfloat r = l + 8;
    GLfloat b = t + 8;

    GLfloat mx = l;
    GLfloat my = t;

    glTexCoord2f(r / w, (h - b) / h);
    glVertex2d(r - mx, b - my);

    glTexCoord2f(l / w, (h - b) / h);
    glVertex2d(l - mx, b - my);

    glTexCoord2f(l / w, (h - t) / h);
    glVertex2d(l - mx, t - my);

    glTexCoord2f(r / w, (h - t) / h);
    glVertex2d(r - mx, t - my);

    glEnd();
}

static void numberxy(int x, int y, int n)
{
    glPushMatrix();
    glTranslatef(x, y, 0);
    if (n < 0)
    {
        subquad(id_numbers, 1);
        glTranslatef(8, 0, 0);
        n = -n;
    }
    int dec = 1000000000;
    while (dec)
    {
        if (n / dec || dec == 1)
        {
            subquad(id_numbers, 4 + n / dec);
            glTranslatef(8, 0, 0);
            n = n % dec;
        }
        dec /= 10;
    }
    glPopMatrix();
}

static void quad(int i, GLfloat cl, GLfloat cr)
{
    glBindTexture(GL_TEXTURE_2D, quads[i].texture);
    glBegin(GL_TRIANGLE_FAN);

    GLfloat w = 16 * 16;
    GLfloat h = 16 * 16;

    GLfloat l = quads[i].l;
    GLfloat t = quads[i].t;
    GLfloat r = quads[i].r + 1;
    GLfloat b = quads[i].b + 1;

    GLfloat mx = l + quads[i].ox;
    GLfloat my = t + quads[i].oy;

    l += cl;
    r -= cr;

    glTexCoord2f(r / w, (h - b) / h);
    glVertex2d(r - mx, b - my);

    glTexCoord2f(l / w, (h - b) / h);
    glVertex2d(l - mx, b - my);

    glTexCoord2f(l / w, (h - t) / h);
    glVertex2d(l - mx, t - my);

    glTexCoord2f(r / w, (h - t) / h);
    glVertex2d(r - mx, t - my);

    glEnd();
}

static void quadxy(int i, GLfloat x, GLfloat y)
{
    glPushMatrix();
    glTranslatef(x, y, 0);
    quad(i, 0, 0);
    glPopMatrix();
}

static void quadxycl(int i, GLfloat x, GLfloat y, GLfloat cl, GLfloat cr)
{
    glPushMatrix();
    glTranslatef(x, y, 0);
    quad(i, cl, cr);
    glPopMatrix();
}

#define MAX_HEIGHT 16
static void render(void)
{
    glClear(GL_COLOR_BUFFER_BIT);

    int ix = isox(scrollx, scrolly);
    int iy = isoy(scrollx, scrolly);

    int px = pixelx(ix, iy) - scrollx;
    int py = pixely(ix, iy) - scrolly;

    if (py > -8)
    {
        iy--;
        px += 16;
        py -= 8;
    }

    while (py < SCREEN_H + MAX_HEIGHT)
    {
        int lix = ix;
        int liy = iy;
        int lpx = px;
        while (px - 16 < SCREEN_W)
        {
            int light = *pos(ix, iy, 7) + *pos(ix, iy, 6);
            if (light)
            {
                glColor3f(light / 255.0, light / 255.0, light / 255.0);
                int c = *pos(ix, iy, 0);
                if (c & 255)
                    quadxy(c & 255, px, py + 24);

                int i;
                for (i = 1; i < 4; i++)
                {
                    c = *pos(ix, iy, i);
                    if (c & 255)
                    {
                        int sx = (char)((c >> 8) & 255);
                        int sy = (char)((c >> 16) & 255);
                        c &= 255;
                        float cl = 0;
                        float cr = 0;
                        if (sx < 0)
                        {
                            int l = quads[c].ox;
                            cl = 16 + sx / 8.0;
                            cl = l - cl;
                            if (cl < 0)
                                cl = 0;

                        }
                        if (sy < 0)
                        {
                            int r = quads[c].r - quads[c].l - quads[c].ox;
                            cr = 16 + sy / 8.0;
                            cr = r - cr;
                            if (cr < 0)
                                cr = 0;
                        }
                        if (sx > 0)
                        {
                            int r = quads[c].r - quads[c].l - quads[c].ox;
                            cr = 16 - sx / 8.0;
                            cr = r - cr;
                            if (cr < 0)
                                cr = 0;
                        }
                        if (sy > 0)
                        {
                            int l = quads[c].ox;
                            cl = 16 - sy / 8.0;
                            cl = l - cl;
                            if (cl < 0)
                                cl = 0;
                        }
                        quadxycl(c, px + sx / 8.0 - sy / 8.0, py + 8 + sx / 16.0 + sy / 16.0,
                            cl, cr);
                    }
                }
            }

            ix++;
            iy--;
            px += 32;
        }
        ix = lix;
        iy = liy;
        px = lpx;
        px -= 16;
        py += 8;
        iy++;
        if (px <= -16)
        {
            px += 32;
            ix++;
            iy--;
        }
    }

    glColor3f(1, 1, 1);

    //cursorxy(mouse_x, mouse_y);
    numberxy(0, 0, babies_left);

    glFlush();
    allegro_gl_flip();
}

int main(void)
{
    srand(time(NULL));
    allegro_init();
    install_allegro_gl();

    int cd = desktop_color_depth();

    allegro_gl_set(AGL_COLOR_DEPTH, cd);
    allegro_gl_set(AGL_SUGGEST, AGL_COLOR_DEPTH);

    allegro_gl_set(AGL_DOUBLEBUFFER, 1);
    allegro_gl_set(AGL_SUGGEST, AGL_DOUBLEBUFFER);

    allegro_gl_set(AGL_SAMPLE_BUFFERS, 0);
    allegro_gl_set(AGL_SUGGEST, AGL_SAMPLE_BUFFERS);

    allegro_gl_set(AGL_SAMPLES, 0);
    allegro_gl_set(AGL_SUGGEST, AGL_SAMPLES);

    allegro_gl_set(AGL_RENDERMETHOD, 1);
    allegro_gl_set(AGL_SUGGEST, AGL_RENDERMETHOD);

    set_color_depth(cd);
    request_refresh_rate(60);
    int mode = windowed ? GFX_OPENGL_WINDOWED : GFX_OPENGL_FULLSCREEN;
    set_gfx_mode(mode, 16 * 40, 16 * 30, 0, 0);

    install_keyboard();
    install_mouse();
    install_timer();
    install_int_ex(ticker, BPS_TO_TIMER(60));
    install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);

    init();

    frames = ticks;
    int rframes = 0;

    while (!key[KEY_ESC] && state != QUIT)
    {
        while (frames <= ticks)
        {
            logic();
            frames++;
        }
        if (rframes < frames)
        {
            render();
            rframes = frames;
        }
        else
            rest(1);
    }

    return 0;
}

END_OF_MAIN()
